# data_process.py
# Author: Pistack
# Date: 2021. 08. 29.
# Email: pistatex@yonsei.ac.kr
# Part of XAS-pre-fit-pack package
# Useful function for doing data processing

import numpy as np
from scipy.interpolate import interp1d


def automate_scaling(A, e_ref_index, e, t,
                     escan_time, tscan_energy, time_zeros=None,
                     escan_data=None, escan_data_eps=None,
                     tscan_data=None, tscan_data_eps=None,
                     warn=False):

    '''
    automate_scaling:
    Automate scale escan, tscan

    
    :param numpy_1d_array A: array of parameter A for each escan
    :param int e_ref: index for reference escan for scaling of escan and tscan
    :param numpy_1d_array e: array of energies in which we measured escan
    :param numpy_1d_array t: array of time delays in which we measured tscan 
    :param numpy_1d_array escan_time: 
     array of time delay at which we measure escan
    :param numpy_1d_array tscan_energy:
     array of energy at which we measure tscan
    :param numpy_1d_array time_zeros:
     array of time zero for every tscan
     (optional, mandatory escan_time[e_ref] < 10 ps)
    :param numpy_nd_array escan_data: data for escan
     (Note. escan data does not contains energy range)
    :param numpy_nd_array escan_data_eps: error for escan data
    :param numpy_nd_array tscan_data: data for tscan
     (Note. tscan data does not contains time delay range)
    :param numpy_nd_array tscan_data_eps: error for tscan
    :param bool warn: whether or not prints warning message [default: False]


    
    :return: scaled_data
     scaled_data['escan'] : scaled data for escan
     scaled_data['escan_eps'] : scaled error for escan
     scaled_data['tscan'] : scaled data for tscan
     scaled_data['tscan_eps'] : scaled error for tscan
    :rtype: dict
    '''

    # init scaled_data
    scaled_data = dict()

    # scaling escan
    scaled_data['escan'] = np.zeros(escan_data.shape)
    scaled_data['escan_eps'] = np.zeros(escan_data_eps.shape)

    A_ref = A[e_ref_index]
    e_ref = escan_time[e_ref_index]

    cond1 = (e_ref < 10)
    
    cond2 = (time_zeros is not None)

    for i in range(escan_data.shape[1]):
        scaled_data['escan'][:, i] = A_ref/A[i]*escan_data[:, i]
        scaled_data['escan_eps'][:, i] = A_ref/A[i]*escan_data_eps[:, i]

    # function for normal scaling procedure
    def do_procedure_normal(time_zeros=None):

        if time_zeros is None:
            time_zeros = np.zeros(tscan_energy.shape[0])

        scaled_data['tscan'] = np.zeros(tscan_data.shape)
        scaled_data['tscan_eps'] = np.zeros(tscan_data_eps.shape)

        flu_ref = interp1d(e,
                           scaled_data['escan'][:, e_ref_index])(tscan_energy)

        for i in range(tscan_energy.shape[0]):

            flu_t = interp1d(t-time_zeros[i], tscan_data[:, i])(e_ref)

            scaled_data['tscan'][:, i] = flu_ref[i]/flu_t*tscan_data[:, i]
            scaled_data['tscan_eps'][:, i] = \
                flu_ref[i]/flu_t*tscan_data_eps[:, i]

        return

    def print_warn(sit):

        if not warn:
            return

        print('Warning !\n')

        if sit == 'fast_delay':
            print(f'Your e_ref: {e_ref} < 10 ps !\n')
            print('In such fast delay, scaling will depend on time zero.\n')
            print('Please watch change of time zeros during fitting !\n')

        return

    def print_err(sit):

        print('Error !\n')

        if sit == 'fast delay without time zero':
            print(f'Your e_ref: {e_ref} < 10 ps !\n')
            print('In such fast delay, scaling will depend on time zero.\n')
            print('However you did not set timezero\n')
            print('Please set timezero first before proceed\n')

        print('Aborting !\n')

        return

    if cond1 and cond2:
        
        do_procedure_normal(time_zeros=time_zeros)

        print_warn('fast_delay')

    elif cond1 and (not cond2):

        print_err('fast delay without time zero')

    else:
        do_procedure_normal(time_zeros=time_zeros)

    return scaled_data


def corr_a_method(e_ref_index, e, t,
                  escan_time, ref_tscan_energy, ref_time_zeros,
                  escan_data=None, escan_data_eps=None,
                  ref_tscan_data=None, ref_tscan_data_eps=None,
                  warn=False):

    '''
    corr_a_method:
    Corrects the scaling of escan scaled with tscan 

    
    :param int e_ref: index of reference escab used for ``A-method``
    :param numpy_1d_array e: array of energies in which we measured escan
    :param numpy_1d_array t: array of time delays in which we measured tscan
    :param numpy_1d_array escan_time: 
     array of time delays at which we measured escan
    :param float ref_tscan_energy: 
     reference energy for repairing scale of escan
    :param float ref_time_zeros:
     time zero for reference tscan
    :param numpy_nd_array escan_data: data for escan
     (Note. escan data does not contains energy range)
    :param numpy_nd_array escan_data_eps: error for escan data 
    :param numpy_1d_array tscan_data: data for reference tscan
    :param numpy_1d_array tscan_data_eps: error for reference tscan 


    
    :return: corrected_data:
     corrected_data['escan'] : corrected data for escan
     corrected_data['escan_eps'] : corrected error for escan
    :rtype: dict
    '''

    # init scaled_data
    scaled_data = dict()

    # scaling escan
    scaled_data['escan'] = np.zeros(escan_data.shape)
    scaled_data['escan_eps'] = np.zeros(escan_data_eps.shape)

    ref_tscan_inv_SN = ref_tscan_data_eps/ref_tscan_data 
    ref_tscan_inv_SN = interp1d(t-ref_time_zeros, 
                                ref_tscan_inv_SN, kind='nearest')(escan_time)
    flu_ref = interp1d(t-ref_time_zeros, ref_tscan_data)(escan_time)
    flu_ref = flu_ref/flu_ref[e_ref_index]
    flu_e = interp1d(e, escan_data, axis=0)(ref_tscan_energy)
    flu_e = flu_e/flu_e[e_ref_index]

    #  Scaling factor obtained by tscan have N/S ambiguity
    for i in range(escan_data.shape[1]):
        cond = np.abs(flu_e[i]-flu_ref[i]) > \
               1.0*ref_tscan_inv_SN[i]*np.max([flu_ref[i], flu_e[i]])
        if cond:
            scaled_data['escan'][:, i] = flu_ref[i]/flu_e[i]*escan_data[:, i]
            scaled_data['escan_eps'][:, i] = \
                flu_ref[i]/flu_e[i]*escan_data_eps[:, i]
        else:
            scaled_data['escan'][:, i] = escan_data[:, i]
            scaled_data['escan_eps'][:, i] = escan_data_eps[:, i]

    return scaled_data



import argparse
import numpy as np
from scipy.interpolate import interp1d 
import matplotlib.pyplot as plt
from ..data_process import automate_scaling
from ..data_process import corr_a_method


def auto_scale():
    description = '''
    auto_scale: Automatic scaling escan and tscan data using
    ``A-method''
    Type ``TRXASprefitpack_info automate_scaling'' to see further
    information
    '''
    epilog = '''
    *Note:
    auto_scale assume each escan have same energy range and
    and each tscan have same time range. 
    Also it assumes energy unit of
    escan data is KeV but assumes energy unit of tscan_energy_file
    is eV. Moreover energy unit of scaled_escan is eV.
    However time unit for tscan data and escan time must be same.
    if you want to know of further information about stage:
    type auto_scale -1 or type ``TRXASprefitpack_info auto_scale
    '''

    stage_m1 = '''
    stage -1: description
    It prints the description about each stages and aborts.
    For every stage except -1, requires 
    prefix, num_of_escan, escan_time, num_of_tscan,
    tscan_energy, time_zeros
    '''
    
    stage0 = '''
    stage 0: init scaling
 
    Additionally requires
    ref_escan_index, tscan_index_to_scale, parm_A

    the program read escan_data from
    prefix_escan_1.txt,...,prefix_num_escan.txt
    and tscan_data from
    prefix_tscan_1.txt,...,prefix_tscan_num_tscan.txt
    Also, it read file for parameter A generated by
    fit static.
    Then it fits scaling of escan_data and 
    tscan_i_1,...,tscan_i_N to escan_e_ref.
    Now it generates 
    prefix_escan_scaled.txt (energy unit: eV)
    prefix_escan_eps_scaled.txt
    prefix_tscan_scaled.txt 
    prefix_tscan_eps_scaled.txt
    prefix_A_ref.txt
    If you do not include tscan j for scaling.
    You can see (j+1) th column of prefix_tscan_scaled.txt and
    j th column of tscan_scaled_eps.txt are filled with zeros.
    '''

    stage1 = '''
    stage 1: Correction

    Additionally requires
    ref_escan_index, tscan_index_to_scale

    Note. 
    ref_escan_index and tscan_index_to_scale must be set to
    same as stage 0
    

    See ``corr_a_method`` to see why correction step is needed
    the program read scaled escan data and tscan data from
    prefix_escan_scaled.txt, prefix_escan_eps_scaled.txt,
    prefix_tscan_scaled.txt, prefix_tscan_eps_scaled.txt,
    and then it corrects scaling of escan using tscan_i_1
    It regenerates prefix_escan_scaled.txt and 
    prefix_escan_eps_scaled.txt
    '''

    stage2 = '''
    stage 2: further scaling

    Additionally requires
    tscan_index_to_scale

    the program read scaled escan data and tscan data from
    prefix_escan_scaled.txt, prefix_escan_eps_scaled.txt,
    prefix_tscan_scaled.txt, prefix_tscan_eps_scaled.txt,
    and then it fits scaling of tscan_i'_1,...,tscan_i'_N'
    to escan_e_ref'. 
    (Prime means i_1,...,i_N and e_ref values are different
     from stage 0)
    Then it regenerates all prefix_*.txt except prefix_A_ref.txt
    
    Note in this stage do not need to give the file name for 
    parameter A.
    '''

    stage3 = '''
    stage 3: sanity check

    In this stage, the program assume, every tscan data are
    scaled to escan data. For sanity check, it gives
    a graph for you.
    '''

    stage4 = '''
    stage 4: Scaling with another tscan data set
    Scale Another set of tscan data using already scaled 
    escan data

    Additionally requires
    tscan_index_to_scale
    
    Before proceed stage 4, you should move 
    prefix_tscan_scaled.txt, prefix_tscan_eps_scaled.txt
    and prefix_tscan_xxx.txt to some backup folder and
    rename your another tscan data set to prefix_tscan_xxx.txt
    Also you should give time_zero and energy for such tscan.

    In this stage, it reads scaled escan data and eps from
    prefix_escan_scaled.txt and prefix_escan_eps_scaled.txt
    Then it reads unscaled tscan data and eps from
    prefix_tscan_*.txt
    Next, it scales tscan datas just like stage 1.
    After stage4 finished you should go to stage 2 and stage 3.
    '''

    stage_description = dict()
    stage_description['-1'] = stage_m1
    stage_description['0'] = stage0
    stage_description['1'] = stage1
    stage_description['2'] = stage2
    stage_description['3'] = stage3
    stage_description['4'] = stage4

    parser = argparse.ArgumentParser(description=description,
                                     epilog=epilog)
    parser.add_argument("stage", choices=['-1', '0', '1', '2', '3', '4'],
                        help="current stage, " +
                        "set stage to -1 get detailed description")
    parser.add_argument("-p", "--prefix",
                        help="prefix for both escan and tscan file, " +
                        "it will read prefix_escan_i.txt " +
                        "and prefix_tscan_j.txt")
    parser.add_argument("-ne", "--num_of_escan",
                        help="the number of escan files",
                        type=int)
    parser.add_argument("-et", "--escan_time",
                        help="filename for escan delay times (unit: ps)")
    parser.add_argument("-re", "--ref_escan_index",
                        help="index of escan used to " +
                        "the reference for scaling",
                        type=int)
    parser.add_argument("-nt", "--num_of_tscan",
                        help="the number of tscan files",
                        type=int)
    parser.add_argument("-te", "--tscan_energy",
                        help="filename for tscan energy (unit: eV)")
    parser.add_argument("-t0", "--time_zeros",
                        help="filename for time zero of each tscan (unit: ps)")
    parser.add_argument("-ti", "--tscan_index_to_scale", nargs='+',
                        type=int,
                        help="tscan index to scale, " +
                        "use blank separation for multiple arguments")
    parser.add_argument("-a", "--parm_A",
                        help="filename for the parameter A obtained from " +
                        "fit_static")

    args = parser.parse_args()
    args_dict = vars(args)
    stage = args.stage
    if stage == '-1':
        for key in stage_description.keys():
            print(stage_description[key])
        return
    else:

        tmp = ['prefix', 'num_of_escan', 'escan_time',
               'num_of_tscan', 'tscan_energy', 'time_zeros']

        for i in tmp:
            if args_dict[i] is None:
                print(f"{i} is required for stage {stage}!\n")
                return

        print(stage_description[stage])

        prefix = args.prefix
        num_of_escan = args.num_of_escan
        escan_time = np.genfromtxt(args.escan_time)
        num_of_tscan = args.num_of_tscan
        tscan_energy = np.genfromtxt(args.tscan_energy)
        time_zeros = np.genfromtxt(args.time_zeros)

    if stage not in ['3', '4']:
        e = np.genfromtxt(f'{prefix}_escan_1.txt')[:, 0]
        e = 1000*e  # KeV to eV

    t = np.genfromtxt(f'{prefix}_tscan_1.txt')[:, 0]
    
    if stage != '3':
        tmp = ['ref_escan_index', 'tscan_index_to_scale']
        for i in tmp:
            if args_dict[i] is None:
                print(f"{i} is required for stage {stage}!")
                return
        
        ref_escan_index = args.ref_escan_index - 1
        
        # select tscan for scaling
        tscan_ind_for_scaling = np.array(args.tscan_index_to_scale)
        tscan_energy_for_scale = tscan_energy[tscan_ind_for_scaling-1]
        time_zeros_for_scale = time_zeros[tscan_ind_for_scaling-1]

        # init variable
        if stage != '4':
            escan_data = np.zeros((e.shape[0], num_of_escan))
            escan_data_eps = np.zeros((e.shape[0], num_of_escan))

        tscan_data = np.zeros((t.shape[0], tscan_ind_for_scaling.shape[0]))
        tscan_data_eps = np.zeros((t.shape[0], tscan_ind_for_scaling.shape[0]))

        for i in range(tscan_ind_for_scaling.shape[0]):
            tmp = f'{prefix}_tscan_{tscan_ind_for_scaling[i]}.txt'
            tscan_data[:, i] = np.genfromtxt(tmp)[:, 1]
            tscan_data_eps[:, i] = np.genfromtxt(tmp)[:, 2]

        if stage in ['0', '4']:
            # init scaled_tscan_data and eps
            scaled_tscan_data = np.zeros((t.shape[0], num_of_tscan+1))
            scaled_tscan_data[:, 0] = t
            scaled_tscan_eps = np.zeros((t.shape[0], num_of_tscan))

            if stage == '4':
                e = np.genfromtxt(f'{prefix}_escan_scaled.txt')[:, 0]
                escan_data = \
                    np.genfromtxt(f'{prefix}_escan_scaled.txt')[:, 1:]
                escan_data_eps = \
                    np.genfromtxt(f'{prefix}_escan_eps_scaled.txt')
                A = np.ones(num_of_escan)
                
            else:
                for i in range(num_of_escan):
                    tmp = f'{prefix}_escan_{i+1}.txt'
                    escan_data[:, i] = np.genfromtxt(tmp)[:, 1]
                    escan_data_eps[:, i] = np.genfromtxt(tmp)[:, 2]
                    tmp = args.parm_A
                    if tmp is None:
                        print(f"parm_A is required for stage {stage}!")
                        return
                    A = np.genfromtxt(args.parm_A)

            scaled_data = automate_scaling(A, ref_escan_index,
                                           e, t,
                                           escan_time, tscan_energy_for_scale,
                                           time_zeros=time_zeros_for_scale,
                                           escan_data=escan_data,
                                           escan_data_eps=escan_data_eps,
                                           tscan_data=tscan_data,
                                           tscan_data_eps=tscan_data_eps,
                                           warn=True)
            
            A_ref = A[ref_escan_index]*np.ones(escan_time.shape[0])
            
            np.savetxt(f'{prefix}_A_ref.txt', A_ref)

        else:
            escan_data = \
                np.genfromtxt(f'{prefix}_escan_scaled.txt')[:, 1:]
            escan_data_eps = \
                np.genfromtxt(f'{prefix}_escan_eps_scaled.txt')

            scaled_tscan_data = \
                np.genfromtxt(f'{prefix}_tscan_scaled.txt')
            scaled_tscan_eps = \
                np.genfromtxt(f'{prefix}_tscan_eps_scaled.txt')
            
            if stage == '1':
                scaled_data = \
                    corr_a_method(ref_escan_index,
                                  e, t,
                                  escan_time,
                                  tscan_energy_for_scale[0],
                                  time_zeros_for_scale[0],
                                  escan_data=escan_data,
                                  escan_data_eps=escan_data_eps,
                                  ref_tscan_data=tscan_data[:, 0],
                                  ref_tscan_data_eps=tscan_data_eps[:, 0])

                escan_data = scaled_data['escan']
                escan_data_eps = scaled_data['escan_eps']
            
            A = np.ones(num_of_escan)
            scaled_data = automate_scaling(A, ref_escan_index,
                                           e, t,
                                           escan_time, tscan_energy_for_scale,
                                           time_zeros=time_zeros_for_scale,
                                           escan_data=escan_data,
                                           escan_data_eps=escan_data_eps,
                                           tscan_data=tscan_data,
                                           tscan_data_eps=tscan_data_eps,
                                           warn=True)

        scaled_tscan_data[:, tscan_ind_for_scaling] = \
            scaled_data['tscan']
        scaled_tscan_eps[:, tscan_ind_for_scaling - 1] = \
            scaled_data['tscan_eps']

        np.savetxt(f'{prefix}_escan_scaled.txt',
                   np.vstack((e, scaled_data['escan'].T)).T)
        np.savetxt(f'{prefix}_escan_eps_scaled.txt',
                   scaled_data['escan_eps'])
        np.savetxt(f'{prefix}_tscan_scaled.txt', scaled_tscan_data)
        np.savetxt(f'{prefix}_tscan_eps_scaled.txt', scaled_tscan_eps)

    else:
        e = np.genfromtxt(f'{prefix}_escan_scaled.txt')[:, 0]
        escan_data = \
            np.genfromtxt(f'{prefix}_escan_scaled.txt')
        escan_data_eps = \
            np.genfromtxt(f'{prefix}_escan_eps_scaled.txt')
        tscan_data = \
            np.genfromtxt(f'{prefix}_tscan_scaled.txt')
        tscan_data_eps = \
            np.genfromtxt(f'{prefix}_tscan_eps_scaled.txt')

        offset_escan = 2*np.amax(abs(escan_data[:, 1:]))
        offset_tscan = 2*np.amax(abs(tscan_data[:, 1:]))
        escan_time = escan_time[escan_time < np.amax(tscan_data[:, 0])]
        num_of_escan = escan_time.shape[0]
        selected_tscan = np.zeros((num_of_tscan, num_of_escan))
        selected_tscan_eps = np.zeros((num_of_tscan, num_of_escan))

        for i in range(num_of_tscan):
            selected_tscan[i, :] = \
                interp1d(t-time_zeros[i], tscan_data[:, i+1])(escan_time)
            selected_tscan_eps[i, :] = \
                interp1d(t-time_zeros[i], tscan_data_eps[:, i],
                         kind='nearest')(escan_time)

        plt.figure(1)
        for i in range(num_of_escan):
            plt.errorbar(e, escan_data[:, i+1]+i*offset_escan,
                         escan_data_eps[:, i],
                         label=f'{escan_time[i]} ps')
            plt.errorbar(e, i*offset_escan*np.ones(e.shape[0]))
        plt.legend()

        plt.figure(2)
        for i in range(num_of_tscan):
            plt.errorbar(t, tscan_data[:, i+1]+i*offset_tscan,
                         tscan_data_eps[:, i],
                         label=f'{tscan_energy[i]} eV')
            plt.errorbar(t, i*offset_tscan*np.ones(t.shape[0]))
        plt.legend()

        plt.figure(3)
        for i in range(num_of_tscan):
            plt.errorbar(t, tscan_data[:, i+1]+i*offset_tscan,
                         tscan_data_eps[:, i],
                         label=f'{tscan_energy[i]} eV')
            plt.errorbar(t, i*offset_tscan*np.ones(t.shape[0]))
        plt.legend()
        plt.xlim(-1, 2)

        for i in range(num_of_escan):
            plt.figure(4+i)
            plt.errorbar(e, escan_data[:, i+1],
                         escan_data_eps[:, i],
                         label=f'{escan_time[i]} ps')
            plt.errorbar(tscan_energy, selected_tscan[:, i],
                         selected_tscan_eps[:, i],
                         linestyle='none', marker='o', mfc='none')

        plt.show()

    return



'''
mathfun:
submodule for the mathematical functions for TRXASprefitpack

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3.
'''

from .mathfun import exp_conv_gau, exp_conv_cauchy
from .mathfun import solve_model, compute_model
from .mathfun import compute_signal_gau, compute_signal_cauchy
from .mathfun import compute_signal_pvoigt
from .mathfun import model_n_comp_conv, fact_anal_exp_conv

__all__ = ['exp_conv_gau', 'exp_conv_cauchy', 'solve_model', 'compute_model',
           'compute_signal_gau', 'compute_signal_cauchy',
           'compute_signal_pvoigt', 'model_n_comp_conv', 'fact_anal_exp_conv']




'''
data_process:

submodule of TRXASprefitpack for data processing

:copyright: 2021 by pistack (Junho Lee).
:license: LGPL3.
'''

from .data_process import automate_scaling, corr_a_method

__all__ = ['automate_scaling', 'corr_a_method']



# mathfun.py
# Author: Pistack
# Date: 2021. 08. 29.
# Email: pistatex@yonsei.ac.kr
# Part of XAS-pre-fit-pack package
# Useful mathematical function for pre-fitting

import numpy as np
import scipy.linalg as LA
from scipy.special import erf, expi


def exp_conv_gau(t, fwhm, k):

    '''
    Compute exponential function convolved with normalized gaussian 
    distribution

    Note.
    We assume temporal pulse of x-ray is normalized gaussian distribution
    
    .. math::

       \sigma = \\frac{fwhm}{2\sqrt{2\log{2}}}


    .. math::

       IRF(t) = \\frac{1}{\sigma \sqrt{2\pi}}\exp\\left(-\\frac{t^2}{2\sigma^2}\\right)


    :param t: time
    :type t: numpy_1d_array
    :param fwhm: full width at half maximum of x-ray temporal pulse
    :type fwhm: float
    :param k: rate constant (inverse of life time)
    :type k: float

    
    :return: convolution of normalized gaussian distribution and exp(-kt)

    .. math::

       \\frac{1}{2}\exp\\left(\\frac{k^2}{2\sigma^2}-kt\\right)\\left(1+{erf}\\left(\\frac{1}{\sqrt{2}}\\left(\\frac{t}{\sigma}-k\sigma\\right)\\right)\\right)


    :rtype: numpy_1d_array
    '''

    sigma = fwhm/(2*np.sqrt(2*np.log(2)))
    return 1/2*np.exp((k*sigma)**2/2.0-k*t) * \
        (1+erf((t/sigma-k*sigma)/np.sqrt(2.0)))


def exp_conv_cauchy(t, fwhm, k):

    '''
    Compute exponential function convolved with normalized cauchy 
    distribution

    Note.
    We assume temporal pulse of x-ray is normalized cauchy distribution

    .. math::
       \\gamma = \\frac{fwhm}{2}

    .. math::
       IRF(t) = \\frac{\\gamma}{\\pi}\\frac{1}{(x-t)^2+\\gamma^2}
    

    :param t: time
    :type t: numpy_1d_array
    :param fwhm: full width at half maximum of x-ray temporal pulse
    :type fwhm: float
    :param k: rate constant (inverse of life time)
    :type k: float

   
    :return: convolution of normalized cauchy distribution and exp(-kt)

    .. math::
       \\frac{\\exp(-kt)}{\\pi}\\Im\\left(\\exp(-ik\\gamma)\cdot\\left(i\\pi - {Ei}\\left(kt+ik\gamma\\right)\\right)\\right)


    :rtype: numpy_1d_array 
    '''

    gamma = fwhm/2
    if k == 0:
        ans = complex(0, 1)*(0.5+1/np.pi*np.arctan(t/gamma))
    else:
        ans = 1/np.pi*np.exp(-k*t)*np.exp(complex(0, -k*gamma)) * \
            (complex(0, np.pi)-expi(k*t+complex(0, k*gamma)))
    
    return ans.imag


def make_A_matrix(t, k):

    A = np.zeros((k.shape[0], t.shape[0]))
    for i in range(k.shape[0]):
        A[i, :] = np.exp(-k[i]*t)
    A = A*np.heaviside(t, 1)
    A[np.isnan(A)] = 0
    
    return A


def make_A_matrix_gau(t, fwhm, k):

    A = np.zeros((k.shape[0], t.shape[0]))
    for i in range(k.shape[0]):
        A[i, :] = exp_conv_gau(t, fwhm, k[i])
    A[np.isnan(A)] = 0
    
    return A


def make_A_matrix_cauchy(t, fwhm, k):

    A = np.zeros((k.shape[0], t.shape[0]))
    for i in range(k.shape[0]):
        A[i, :] = exp_conv_cauchy(t, fwhm, k[i])
    A[np.isnan(A)] = 0
    A[np.isinf(A)] = 0
    
    return A


def make_A_matrix_pvoigt(t, fwhm_G, fwhm_L, eta, k):
    return (1-eta)*make_A_matrix_gau(t, fwhm_G, k) + \
        eta*make_A_matrix_cauchy(t, fwhm_L, k)


def solve_model(equation, y0):

    '''
    Solve system of first order rate equation

    
    :param equation: matrix corresponding to model
    :type equation: numpy_nd_array
    :param y0: initial condition
    :type y0: numpy_1d_array

    
    :return: eigenvalue of equation
    :rtype: numpy_1d_array
    :return: eigenvectors for equation
    :rtype: numpy_nd_array
    :return: coefficient where y_0 = Vc
    :rtype: numpy_1d_array
    '''

    eigval, V = LA.eig(equation)
    c = LA.solve(V, y0)

    return eigval.real, V, c


def compute_model(t, eigval, V, c):

    '''
    Compute solution of the system of rate equations solved by solve_model
    Note: eigval, V, c should be obtained from solve_model

    
    :param t: time
    :type t: numpy_1d_array
    :param eigval: eigenvalue for equation
    :type eigval: numpy_1d_array
    :param V: eigenvectors for equation
    :type V: numpy_nd_array
    :param c: coefficient
    :type c: numpy_1d_array

    
    :return: solution of rate equation
    :rtype: numpy_nd_array
    '''

    A = make_A_matrix(t, -eigval)
    y = (c * V) @ A
    return y


def compute_signal_gau(t, fwhm, eigval, V, c):

    '''
    Compute solution of the system of rate equations solved by solve_model
    convolved with normalized gaussian distribution
    Note: eigval, V, c should be obtained from solve_model

    
    :param t: time
    :type t: numpy_1d_array
    :param fwhm: full width at half maximum of x-ray temporal pulse
    :type fwhm: float
    :param eigval: eigenvalue for equation
    :type eigval: numpy_1d_array
    :param V: eigenvectors for equation
    :type V: numpy_nd_array
    :param c: coefficient
    :type c: numpy_1d_array
  

    :return: 
     solution of rate equation convolved with normalized gaussian distribution
    :rtype: numpy_nd_array 
    '''

    A = make_A_matrix_gau(t, fwhm, -eigval)
    y_signal = (c * V) @ A
    return y_signal


def compute_signal_cauchy(t, fwhm, eigval, V, c):

    '''
    Compute solution of the system of rate equations solved by solve_model
    convolved with normalized cauchy distribution
    Note: eigval, V, c should be obtained from solve_model

    
    :param t: time
    :type t: numpy_1d_array
    :param fwhm: full width at half maximum of x-ray temporal pulse
    :type fwhm: float
    :param eigval: eigenvalue for equation
    :type eigval: numpy_1d_array
    :param V: eigenvectors for equation
    :type V: numpy_nd_array
    :param c: coefficient
    :type c: numpy_1d_array
  

    :return: 
     solution of rate equation convolved with normalized cachy distribution
    :rtype: numpy_nd_array 
    '''

    A = make_A_matrix_cauchy(t, fwhm, -eigval)
    y_signal = (c * V) @ A
    return y_signal


def compute_signal_pvoigt(t, fwhm_G, fwhm_L, eta, eigval, V, c):

    '''
    Compute solution of the system of rate equations solved by solve_model
    convolved with normalized pseudo voigt profile
    (:math:`pvoigt = (1-\\eta) G(t) + \\eta L(t)`,
    G(t): stands for normalized gaussian
    L(t): stands for normalized cauchy(lorenzian) distribution)

    Note: eigval, V, c should be obtained from solve_model

    :param t: time
    :type t: numpy_1d_array
    :param fwhm_G: 
     full width at half maximum of x-ray temporal pulse (gaussian part)
    :type fwhm_G: float
    :param fwhm_L:
     full width at half maximum of x-ray temporal pulse (lorenzian part)
    :type fwhm_L: float
    :param float eta: mixing parameter

    :math:`(0 < \\eta < 1)`

    :type eta: float
    :param eigval: eigenvalue for equation
    :type eigval: numpy_1d_array
    :param V: eigenvectors for equation
    :type V: numpy_nd_array
    :param c: coefficient
    :type c: numpy_1d_array

    :return: 
     solution of rate equation convolved with normalized pseudo voigt profile
    :rtype: numpy_nd_array
    '''

    A = make_A_matrix_pvoigt(t, fwhm_G, fwhm_L, eta, -eigval)
    y_signal = (c * V) @ A
    return y_signal


def model_n_comp_conv(t, fwhm, tau, c, base=True, irf='g', eta=None):

    '''
    model for n component fitting
    n exponential function convolved with irf
    irf
    'g': normalized gaussian distribution

    'c': normalized cauchy distribution

    'pv': pseudo voigt profile :math:`(1-\\eta)g + \\eta c`


    :param numpy_1d_array t: time
    :param numpy_1d_array fwhm: fwhm of X-ray temporal pulse 
     if irf == 'g' or 'c' then
        fwhm = [fwhm]
     if irf == 'pv' then
        fwhm = [fwhm_G, fwhm_L]
    :param numpy_1d_array tau: life time for each component
    :param numpy_1d_array c: coefficient
     (num_comp+1,) if base=True
     (num_comp,)   if base=False
    :param base: whether or not include baseline [default: True]
    :type base: bool, optional
    :param irf: shape of instrumental response function [default: g]
     'g': normalized gaussian distribution
     'c': normalized cauchy distribution
     'pv': pseudo voigt profile :math:`(1-\\eta)g + \\eta c`

    :type irf: string, optional
    :param eta: mixing parameter for pseudo voigt profile 
     (only needed for pseudo voigt profile, Default value is guessed according to Journal of Applied Crystallography. 33 (6): 1311–1316.)
    :type eta: float, optional
    
    :return: fit
    :rtype: numpy_1d_array
    '''

    k = 1/tau
    if base:
        k = np.concatenate((k, np.array([0])))

    if irf == 'g':
        A = make_A_matrix_gau(t, fwhm[0], k)
    elif irf == 'c':
        A = make_A_matrix_cauchy(t, fwhm[0], k)
    elif irf == 'pv':
        if eta is None:
            f = fwhm[0]**5+2.69269*fwhm[0]**4*fwhm[1] + \
                2.42843*fwhm[0]**3*fwhm[1]**2 + \
                4.47163*fwhm[0]**2*fwhm[1]**3 + \
                0.07842*fwhm[0]*fwhm[1]**4 + \
                fwhm[1]**5
            f = f**(1/5)
            x = fwhm[1]/f
            eta = 1.36603*x-0.47719*x**2+0.11116*x**3
        A = make_A_matrix_pvoigt(t, fwhm[0], fwhm[1],
                                 eta, k)

    y = c@A

    return y


def fact_anal_exp_conv(t, fwhm, tau, irf='g', eta=None,
                       data=None, eps=None, base=True):
    
    '''
    Estimate the best coefficiets when fwhm and tau are given

    Before fitting with Escan data, you want fit your model to tscan data
    to know how many component needed for well description of tscan data
    To do this you have good initial guess for not only life time of
    each component but also coefficients.
    For this, It will solve linear least square problem

    :param numpy_1d_array t: time
    :param numpy_1d_array fwhm: fwhm of X-ray temporal pulse 
     if irf == 'g' or 'c' then
        fwhm = [fwhm]
     if irf == 'pv' then
        fwhm = [fwhm_G, fwhm_L]
    :param numpy_1d_array tau: life time for each component
    :param irf: shape of instrumental response function [default: g]
     'g': normalized gaussian distribution
     'c': normalized cauchy distribution
     'pv': pseudo voigt profile :math:`(1-\\eta)g + \\eta c`
    :type irf: string, optional
    :param eta: mixing parameter for pseudo voigt profile 
     (only needed for pseudo voigt profile, Default value is guessed according to Journal of Applied Crystallography. 33 (6): 1311–1316.)
    :type eta: float, optional
    :param numpy_1d_array data: tscan data
    :param numpy_1d_array eps: error for tscan data 
    :param base: whether or not include baseline [default: True]
    :type base: bool, optional
    

    :return: best coefficient for given fwhm, tau
     (num_comp+1,) , if base=True
     (num_comp,)   , otherwise
    :rtype: numpy_1d_array
    '''

    k = 1/tau
    if base:
        k = np.concatenate((k, np.array([0])))

    if irf == 'g':
        A = make_A_matrix_gau(t, fwhm[0], k)
    elif irf == 'c':
        A = make_A_matrix_cauchy(t, fwhm[0], k)
    elif irf == 'pv':
        if eta is None:
            f = fwhm[0]**5+2.69269*fwhm[0]**4*fwhm[1] + \
                2.42843*fwhm[0]**3*fwhm[1]**2 + \
                4.47163*fwhm[0]**2*fwhm[1]**3 + \
                0.07842*fwhm[0]*fwhm[1]**4 + \
                fwhm[1]**5
            f = f**(1/5)
            x = fwhm[1]/f
            eta = 1.36603*x-0.47719*x**2+0.11116*x**3
        A = make_A_matrix_pvoigt(t, fwhm[0], fwhm[1], eta, k)

    if eps is not None:
        y = data/eps
        for i in range(k.shape[0]):
            A[i, :] = A[i, :]/eps
    else:
        y = data
    c, _, _, _ = LA.lstsq(A.T, y)

    return c



'''
tools:
submodule for useful main functions

:copyright: (c) 2021 by pistack (Junho Lee).
:license: LGPL3
'''

from .TRXASprefitpack_info import TRXASprefitpack_info
from .fit_static import fit_static
from .fit_tscan import fit_tscan
from .auto_scale import auto_scale
from .broadening import broadening



# info.py:
# copyright: 2021 by pistatex (Junho Lee).
# license: LGPL3.
# Automately generated by gen_doc.py

class docs:

    ''' class for store the documentation for TRXASprefitpack '''
    
    def __init__(self):
        self.info = dict()
        return
    
    def appnd(self, name, txt):
        self.info[name] = txt
        return

__info__ = docs()

doc = dict()

doc['version'] = ' 0.4.1' 
doc['description'] = '''
TRXASprefitpack: package for TRXAS pre-fitting process which aims for the
first order dynamics
version:  0.4.1

                           numpy, scipy
                                |
                                |
                                V
             ***************TRXASprefitpack***************
             | thy             -- gen_theory_data        |
             | data_process    -- automate_scaling       |
             |                 -- corr_a_method          |
             | mathfun         -- exp_conv_gau           |
             |                 -- exp_conv_cauchy        |
             |                 -- solve_model            |
	     |                 -- compute_model          |
	     |                 -- compute_signal_gau     |
             |                 -- compute_signal_cauchy  |
             |                 -- compute_signal_pvoigt  |_________
             |                 -- model_n_comp_conv      |         |
             |                 -- fact_anal_exp_conv     |         |
             | doc             -- info                   |         |
             *********************************************         |
                                |                                  |
                                |                                  |
                                |                                  | lmfit
				V                                  | matplotlib
             ****************script*******************             |
             | TRXASprefitpack_info  (exe)           |             |
             ****************script*******************             |         
             | broadening           (exe)            |             |
             | auto_scale           (exe)            |             |
             | fit_static           (exe)            |             |
	     | fit_tscan            (exe)            | <------------
             *****************************************


If you want know any information about function defined in TRXASprefitpack
type TRXASprefitpack_info func_name
'''
 
doc['licence'] = '''
TRXASprefitpack: package for TRXAS pre fitting process

Copyright (C) 2021 pistack (Junho Lee, email: pistatex@yonsei.ac.kr)
 
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.
 
You should have received a copy of the GNU Lesser General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''
 
doc['lgpl-3.0'] = '''
                  GNU LESSER GENERAL PUBLIC LICENSE
                       Version 3, 29 June 2007

 Copyright (C) 2007 Free Software Foundation, Inc. <https://fsf.org/>
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.


  This version of the GNU Lesser General Public License incorporates
the terms and conditions of version 3 of the GNU General Public
License, supplemented by the additional permissions listed below.

  0. Additional Definitions.

  As used herein, "this License" refers to version 3 of the GNU Lesser
General Public License, and the "GNU GPL" refers to version 3 of the GNU
General Public License.

  "The Library" refers to a covered work governed by this License,
other than an Application or a Combined Work as defined below.

  An "Application" is any work that makes use of an interface provided
by the Library, but which is not otherwise based on the Library.
Defining a subclass of a class defined by the Library is deemed a mode
of using an interface provided by the Library.

  A "Combined Work" is a work produced by combining or linking an
Application with the Library.  The particular version of the Library
with which the Combined Work was made is also called the "Linked
Version".

  The "Minimal Corresponding Source" for a Combined Work means the
Corresponding Source for the Combined Work, excluding any source code
for portions of the Combined Work that, considered in isolation, are
based on the Application, and not on the Linked Version.

  The "Corresponding Application Code" for a Combined Work means the
object code and/or source code for the Application, including any data
and utility programs needed for reproducing the Combined Work from the
Application, but excluding the System Libraries of the Combined Work.

  1. Exception to Section 3 of the GNU GPL.

  You may convey a covered work under sections 3 and 4 of this License
without being bound by section 3 of the GNU GPL.

  2. Conveying Modified Versions.

  If you modify a copy of the Library, and, in your modifications, a
facility refers to a function or data to be supplied by an Application
that uses the facility (other than as an argument passed when the
facility is invoked), then you may convey a copy of the modified
version:

   a) under this License, provided that you make a good faith effort to
   ensure that, in the event an Application does not supply the
   function or data, the facility still operates, and performs
   whatever part of its purpose remains meaningful, or

   b) under the GNU GPL, with none of the additional permissions of
   this License applicable to that copy.

  3. Object Code Incorporating Material from Library Header Files.

  The object code form of an Application may incorporate material from
a header file that is part of the Library.  You may convey such object
code under terms of your choice, provided that, if the incorporated
material is not limited to numerical parameters, data structure
layouts and accessors, or small macros, inline functions and templates
(ten or fewer lines in length), you do both of the following:

   a) Give prominent notice with each copy of the object code that the
   Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the object code with a copy of the GNU GPL and this license
   document.

  4. Combined Works.

  You may convey a Combined Work under terms of your choice that,
taken together, effectively do not restrict modification of the
portions of the Library contained in the Combined Work and reverse
engineering for debugging such modifications, if you also do each of
the following:

   a) Give prominent notice with each copy of the Combined Work that
   the Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the Combined Work with a copy of the GNU GPL and this license
   document.

   c) For a Combined Work that displays copyright notices during
   execution, include the copyright notice for the Library among
   these notices, as well as a reference directing the user to the
   copies of the GNU GPL and this license document.

   d) Do one of the following:

       0) Convey the Minimal Corresponding Source under the terms of this
       License, and the Corresponding Application Code in a form
       suitable for, and under terms that permit, the user to
       recombine or relink the Application with a modified version of
       the Linked Version to produce a modified Combined Work, in the
       manner specified by section 6 of the GNU GPL for conveying
       Corresponding Source.

       1) Use a suitable shared library mechanism for linking with the
       Library.  A suitable mechanism is one that (a) uses at run time
       a copy of the Library already present on the user's computer
       system, and (b) will operate properly with a modified version
       of the Library that is interface-compatible with the Linked
       Version.

   e) Provide Installation Information, but only if you would otherwise
   be required to provide such information under section 6 of the
   GNU GPL, and only to the extent that such information is
   necessary to install and execute a modified version of the
   Combined Work produced by recombining or relinking the
   Application with a modified version of the Linked Version. (If
   you use option 4d0, the Installation Information must accompany
   the Minimal Corresponding Source and Corresponding Application
   Code. If you use option 4d1, you must provide the Installation
   Information in the manner specified by section 6 of the GNU GPL
   for conveying Corresponding Source.)

  5. Combined Libraries.

  You may place library facilities that are a work based on the
Library side by side in a single library together with other library
facilities that are not Applications and are not covered by this
License, and convey such a combined library under terms of your
choice, if you do both of the following:

   a) Accompany the combined library with a copy of the same work based
   on the Library, uncombined with any other library facilities,
   conveyed under the terms of this License.

   b) Give prominent notice with the combined library that part of it
   is a work based on the Library, and explaining where to find the
   accompanying uncombined form of the same work.

  6. Revised Versions of the GNU Lesser General Public License.

  The Free Software Foundation may publish revised and/or new versions
of the GNU Lesser General Public License from time to time. Such new
versions will be similar in spirit to the present version, but may
differ in detail to address new problems or concerns.

  Each version is given a distinguishing version number. If the
Library as you received it specifies that a certain numbered version
of the GNU Lesser General Public License "or any later version"
applies to it, you have the option of following the terms and
conditions either of that published version or of any later version
published by the Free Software Foundation. If the Library as you
received it does not specify a version number of the GNU Lesser
General Public License, you may choose any version of the GNU Lesser
General Public License ever published by the Free Software Foundation.

  If the Library as you received it specifies that a proxy can decide
whether future versions of the GNU Lesser General Public License shall
apply, that proxy's public statement of acceptance of any version is
permanent authorization for you to choose that version for the
Library.
'''
 
doc['gen_theory_data'] = '''
gen_theory_data(e, peaks, A, fwhm_G, fwhm_L, peak_shift, out=None)
    
   voigt broadening theoretically calculated lineshape spectrum if out
   is not none: It will make out_thy.txt: txt file for rescaled and
   boroadend calc spectrum out_thy_stk.txt: txt file for rescaled and
   shifted calc peaks

   Parameters:
      * **e** (*numpy_1d_array*) -- energy (unit: eV)

      * **A** (*float*) -- scaling parameter

      * **fwhm_G** (*float*) -- full width at half maximum of gaussian
        shape (unit: eV)

      * **fwhm_L** (*float*) -- full width at half maximum of
        lorenzian shape (unit: eV)

      * **peak_shift** (*float*) -- discrepency of peak position
        between expt data and theoretically  broadened spectrum

      * **out** (*string*) -- prefix for output txt file [optional]

   Returns:
      voigt broadened calc spectrum

   Return type:
      numpy_1d_array
''' 
 
doc['automate_scaling'] = '''    
    automate_scaling(A, e_ref_index, e, t,
                     escan_time, tscan_energy, time_zeros=None,
                     exotic_t=None, exotic_tscan_energy=None,
                     exotic_e_ref_index=None,
                     escan_data=None, escan_data_eps=None,
                     tscan_data=None, tscan_data_eps=None,
                     exotic_tscan_data=None, exotic_tscan_data_eps=None)
                     
    Automate scale escan, tscan and some exotic tscan
    If you scale tscan with fast time delay ( < 10 ps)
    time zero will change scaling factor. So, beware of the
    change of time zero.
    However, scaling parameter for tscan w.r.t. escan (r_t_i)
    may recover such effect. So, I think that for not so early time delay
    (1 ps ~ 10 ps) region time_zero shifted about +/- IRF ~ 300 fs
    neglectable.
    If escan_time[e_ref_index] < 10 ps
    I will print warning message for you.

    Automate scaling procedure

    0. Fit static spectrum with fit_static.py
       Carefully watch graphs.

    1. Scaling escan_data using A-method

    2-1. If escan_time[e_ref_index] := e_ref > 10 ps,
         then take difference flu spectrum at tscan energy
         in e_ref delay escan.
         Next take difference flu spectrum at e_ref delay in tscan.
         Now fit diff flu spec at e_ref delay in tscan to
         diff flu spec at tscan_energy in e_ref delay escan.

    2-2. If e_ref < 10 ps, you must set time_zeros for all tscan.
         Then do procedure simuliar to 2-2.

    3. Watch changes in timezero during fitting.
       If timezero changes a lot in fitting, consider simultaneous
       scaling (i.e. fitting and scaling at once.)

    4. After do automate_scaling, you need to correct scaling
       due to the prossiblity of inconsistent laser overlap.
   
    5. Correct and do automate_scaling again with corrected escan and
       A = A_ref*np.ones(A.shape).


    Warning and Error case

    Warning (no harm, just for warn)
    
    1. early delay (e_ref < 10)

    Error (it aborts procedure)

    1. early delay without time zero : e_ref < 10 but time zero is not set.

   automate_scaling: Automate scale escan, tscan

   Parameters:
      * **A** (*numpy_1d_array*) -- array of parameter A for each
        escan

      * **e_ref** (*int*) -- index for reference escan for scaling of
        escan and tscan

      * **e** (*numpy_1d_array*) -- array of energies in which we
        measured escan

      * **t** (*numpy_1d_array*) -- array of time delays in which we
        measured tscan

      * **escan_time** (*numpy_1d_array*) -- array of time delay at
        which we measure escan

      * **tscan_energy** (*numpy_1d_array*) -- array of energy at
        which we measure tscan

      * **time_zeros** (*numpy_1d_array*) -- array of time zero for
        every tscan (optional, mandatory escan_time[e_ref] < 10 ps)

      * **escan_data** (*numpy_nd_array*) -- data for escan (Note.
        escan data does not contains energy range)

      * **escan_data_eps** (*numpy_nd_array*) -- error for escan data

      * **tscan_data** (*numpy_nd_array*) -- data for tscan (Note.
        tscan data does not contains time delay range)

      * **tscan_data_eps** (*numpy_nd_array*) -- error for tscan

      * **warn** (*bool*) -- whether or not prints warning message
        [default: False]

   Returns:
      scaled_data scaled_data['escan'] : scaled data for escan
      scaled_data['escan_eps'] : scaled error for escan
      scaled_data['tscan'] : scaled data for tscan
      scaled_data['tscan_eps'] : scaled error for tscan

   Return type:
      dict
'''
 
doc['corr_a_method'] = '''
Due to inconsistent of laser overlap during experiment
We may need to correct ``A-method`` scaled escan data 
corr_a_method correct this data using reference tscan

corr_a_method(e_ref_index, e, t,
                     escan_time, ref_tscan_energy, ref_time_zeros,
                     escan_data=None, escan_data_eps=None,
                     ref_tscan_data=None, ref_tscan_data_eps=None,
                     warn=False):

   corr_a_method: Corrects the scaling of escan scaled with tscan

   Parameters:
      * **e_ref** (*int*) -- index of reference escab used for
        "A-method"

      * **e** (*numpy_1d_array*) -- array of energies in which we
        measured escan

      * **t** (*numpy_1d_array*) -- array of time delays in which we
        measured tscan

      * **escan_time** (*numpy_1d_array*) -- array of time delays at
        which we measured escan

      * **ref_tscan_energy** (*float*) -- reference energy for
        repairing scale of escan

      * **ref_time_zeros** (*float*) -- time zero for reference tscan

      * **escan_data** (*numpy_nd_array*) -- data for escan (Note.
        escan data does not contains energy range)

      * **escan_data_eps** (*numpy_nd_array*) -- error for escan data

      * **tscan_data** (*numpy_1d_array*) -- data for reference tscan

      * **tscan_data_eps** (*numpy_1d_array*) -- error for reference
        tscan

   Returns:
      corrected_data: corrected_data['escan'] : corrected data for
      escan corrected_data['escan_eps'] : corrected error for escan

   Return type:
      dict
'''
 
doc['exp_conv_gau'] = '''
exp_conv_gau(t, fwhm, k):

   Compute exponential function convolved with normalized gaussian
   distribution

   Note. We assume temporal pulse of x-ray is normalized gaussian
   distribution

      \sigma = \frac{fwhm}{2\sqrt{2\log{2}}}

      IRF(t) = \frac{1}{\sigma
      \sqrt{2\pi}}\exp\left(-\frac{t^2}{2\sigma^2}\right)

   Parameters:
      * **t** (*numpy_1d_array*) -- time

      * **fwhm** (*float*) -- full width at half maximum of x-ray
        temporal pulse

      * **k** (*float*) -- rate constant (inverse of life time)

   Returns:
      convolution of normalized gaussian distribution and exp(-kt)

      \frac{1}{2}\exp\left(\frac{k^2}{2\sigma^2}-kt\right)\left(1+{er
      f}\left(\frac{1}{\sqrt{2}}\left(\frac{t}{\sigma}-k\sigma\right)
      \right)\right)

   Return type:
      numpy_1d_array    
'''
 
doc['exp_conv_cauchy'] = '''
exp_conv_cauchy(t, fwhm, k):

   Compute exponential function convolved with normalized cauchy
   distribution

   Note. We assume temporal pulse of x-ray is normalized cauchy
   distribution

      \gamma = \frac{fwhm}{2}

      IRF(t) = \frac{\gamma}{\pi}\frac{1}{(x-t)^2+\gamma^2}

   Parameters:
      * **t** (*numpy_1d_array*) -- time

      * **fwhm** (*float*) -- full width at half maximum of x-ray
        temporal pulse

      * **k** (*float*) -- rate constant (inverse of life time)

   Returns:
      convolution of normalized cauchy distribution and exp(-kt)

      \frac{\exp(-kt)}{\pi}\Im\left(\exp(-ik\gamma)\cdot\left(i\pi -
      {Ei}\left(kt+ik\gamma\right)\right)\right)

   Return type:
      numpy_1d_array
'''
 
doc['solve_model'] = '''
 solve_model(equation, y0):

   Solve system of first order rate equation

   Parameters:
      * **equation** (*numpy_nd_array*) -- matrix corresponding to
        model

      * **y0** (*numpy_1d_array*) -- initial condition

   Returns:
      eigenvalue of equation

   Return type:
      numpy_1d_array

   Returns:
      eigenvectors for equation

   Return type:
      numpy_nd_array

   Returns:
      coefficient where y_0 = Vc

   Return type:
      numpy_1d_array
 '''
    
 
doc['compute_model'] = '''
compute_model(t, eigval, V, c):

   Compute solution of the system of rate equations solved by
   solve_model Note: eigval, V, c should be obtained from solve_model

   Parameters:
      * **t** (*numpy_1d_array*) -- time

      * **eigval** (*numpy_1d_array*) -- eigenvalue for equation

      * **V** (*numpy_nd_array*) -- eigenvectors for equation

      * **c** (*numpy_1d_array*) -- coefficient

   Returns:
      solution of rate equation

   Return type:
      numpy_nd_array
'''
 
doc['compute_signal_gau'] = '''
compute_signal_gau(t, fwhm, eigval, V, c):

   Compute solution of the system of rate equations solved by
   solve_model convolved with normalized gaussian distribution Note:
   eigval, V, c should be obtained from solve_model

   Parameters:
      * **t** (*numpy_1d_array*) -- time

      * **fwhm** (*float*) -- full width at half maximum of x-ray
        temporal pulse

      * **eigval** (*numpy_1d_array*) -- eigenvalue for equation

      * **V** (*numpy_nd_array*) -- eigenvectors for equation

      * **c** (*numpy_1d_array*) -- coefficient

   Returns:
      solution of rate equation convolved with normalized gaussian
      distribution

   Return type:
      numpy_nd_array
'''
 
doc['compute_signal_cauchy'] = '''
compute_signal_cauchy(t, fwhm, eigval, V, c):

   Compute solution of the system of rate equations solved by
   solve_model convolved with normalized cauchy distribution Note:
   eigval, V, c should be obtained from solve_model

   Parameters:
      * **t** (*numpy_1d_array*) -- time

      * **fwhm** (*float*) -- full width at half maximum of x-ray
        temporal pulse

      * **eigval** (*numpy_1d_array*) -- eigenvalue for equation

      * **V** (*numpy_nd_array*) -- eigenvectors for equation

      * **c** (*numpy_1d_array*) -- coefficient

   Returns:
      solution of rate equation convolved with normalized cachy
      distribution

   Return type:
      numpy_nd_array
'''
 
doc['compute_signal_pvoigt'] = '''
compute_signal_pvoigt(t, fwhm_G, fwhm_L, eta, eigval, V, c):

   Compute solution of the system of rate equations solved by
   solve_model convolved with normalized pseudo voigt profile (pvoigt
   = (1-\eta) G(t) + \eta L(t), G(t): stands for normalized gaussian
   L(t): stands for normalized cauchy(lorenzian) distribution)

   Note: eigval, V, c should be obtained from solve_model

   Parameters:
      * **t** (*numpy_1d_array*) -- time

      * **fwhm_G** (*float*) -- full width at half maximum of x-ray
        temporal pulse (gaussian part)

      * **fwhm_L** (*float*) -- full width at half maximum of x-ray
        temporal pulse (lorenzian part)

      * **eta** (*float*) -- mixing parameter

   (0 < \eta < 1)

   Parameters:
      * **eigval** (*numpy_1d_array*) -- eigenvalue for equation

      * **V** (*numpy_nd_array*) -- eigenvectors for equation

      * **c** (*numpy_1d_array*) -- coefficient

   Returns:
      solution of rate equation convolved with normalized pseudo voigt
      profile

   Return type:
      numpy_nd_array
'''
 
doc['model_n_comp_conv'] = '''
model_n_comp_conv(t, fwhm, tau, c, base=True, irf='g', eta=None):

   model for n component fitting n exponential function convolved with
   irf 'g': normalized gaussian distribution

   'c': normalized cauchy distribution

   'pv': pseudo voigt profile (1-\eta)g + \eta c

   Parameters:
      * **t** (*numpy_1d_array*) -- time

      * **fwhm** (*numpy_1d_array*) --

        fwhm of X-ray temporal pulse  if irf == 'g' or 'c' then

           fwhm = [fwhm]

        if irf == 'pv' then
           fwhm = [fwhm_G, fwhm_L]

      * **tau** (*numpy_1d_array*) -- life time for each component

      * **c** (*numpy_1d_array*) -- coefficient (num_comp+1,) if
        base=True (num_comp,)   if base=False

      * **base** (*bool**, **optional*) -- whether or not include
        baseline [default: True]

      * **irf** (*string**, **optional*) -- shape of instrumental
        response function [default: g] 'g': normalized gaussian
        distribution 'c': normalized cauchy distribution 'pv': pseudo
        voigt profile (1-\eta)g + \eta c

      * **eta** (*float**, **optional*) -- mixing parameter for pseudo
        voigt profile  (only needed for pseudo voigt profile, Default
        value is guessed according to Journal of Applied
        Crystallography. 33 (6): 1311–1316.)

   Returns:
      fit

   Return type:
      numpy_1d_array

'''
 
doc['fact_anal_exp_conv'] = '''
fact_anal_exp_conv(t, fwhm, tau, irf='g', eta=None, 
                   data=None, eps=None, base=True):

   Estimate the best coefficiets when fwhm and tau are given

   Before fitting with Escan data, you want fit your model to tscan
   data to know how many component needed for well description of
   tscan data To do this you have good initial guess for not only life
   time of each component but also coefficients. For this, It will
   solve linear least square problem

   Parameters:
      * **t** (*numpy_1d_array*) -- time

      * **fwhm** (*numpy_1d_array*) --

        fwhm of X-ray temporal pulse  if irf == 'g' or 'c' then

           fwhm = [fwhm]

        if irf == 'pv' then
           fwhm = [fwhm_G, fwhm_L]

      * **tau** (*numpy_1d_array*) -- life time for each component

      * **irf** (*string**, **optional*) -- shape of instrumental
        response function [default: g] 'g': normalized gaussian
        distribution 'c': normalized cauchy distribution 'pv': pseudo
        voigt profile (1-\eta)g + \eta c

      * **eta** (*float**, **optional*) -- mixing parameter for pseudo
        voigt profile  (only needed for pseudo voigt profile, Default
        value is guessed according to Journal of Applied
        Crystallography. 33 (6): 1311–1316.)

      * **data** (*numpy_1d_array*) -- tscan data

      * **eps** (*numpy_1d_array*) -- error for tscan data

      * **base** (*bool**, **optional*) -- whether or not include
        baseline [default: True]

   Returns:
      best coefficient for given fwhm, tau (num_comp+1,) , if
      base=True (num_comp,)   , otherwise

   Return type:
      numpy_1d_array
'''
 
doc['fit_static'] = '''
fit_static
**********

fitting static spectrum with theoretically calculated line spectrum
broadened by spectral line shape

Note:

  Currently, it uses linear baseline.

Note:

  energy unit of measured static spectrum must be KeV and energy unit
  of calc static spectrum must be eV

* usage: fit_static [-h] [-ls {v,g,l}] [-o OUT] prefix num_scan
  peak_name

* positional arguments:

  * prefix                prefix for experimental static peak files It
    will read prefix_i.txt files

  * num_scan              the number of static peak scan files

  * peak_name             filename for theoretical line shape spectrum

* optional arguments:

  * -h, --help            show this help message and exit

  * -ls {v,g,l}, --line_shape {v,g,l} line shape of spectrum v: voigt
    profile g: gaussian shape l: lorenzian shape

  * -o OUT, --out OUT     prefix for output files
'''
 
doc['fit_tscan'] = '''
fit_tscan
*********

fitting tscan data using sum of exponential decay covolved with
gaussian/cauchy(lorenzian)/pseudo voigt irf function. It uses
"fact_anal_exp_conv" to determine best c_i's where timezero, fwhm, and
time constants are given. So, to use this script what you need to give
are only timezero, fwhm, and time constants

Note:

  * If you set shape of irf to pseudo voigt (pv), then you should
    provide two full width at half maximum value for gaussian and
    cauchy parts, respectively.

  * If you did not set tau then it assume you finds the timezero of
    this scan. So, --no_base option is discouraged.

* usage: fit_tscan    [-h] [--irf {g,c,pv}] [--fwhm_G FWHM_G]
  [--fwhm_L FWHM_L] [-t0 TIME_ZEROS [TIME_ZEROS ...]] [-t0f
  TIME_ZEROS_FILE] [--tau [TAU [TAU ...]]] [--no_base] [-o OUT] prefix

* positional arguments:

  * prefix                prefix for tscan files It will read
    prefix_i.txt

* optional arguments:

  * -h, --help            show this help message and exit

  * --irf {g,c,pv}        shape of instrument response function g:
    gaussian distribution c: cauchy distribution pv: pseudo voigt
    profile, linear combination of gaussian distribution and cauchy
    distribution pv = eta*c+(1-eta)*g the mixing parameter is guessed
    according to Journal of Applied Crystallography. 33 (6):
    1311–1316.

  * --fwhm_G FWHM_G       full width at half maximum for gaussian
    shape It should not used when you set cauchy irf function

  * --fwhm_L FWHM_L       full width at half maximum for cauchy shape
    It should not used when you did not set irf or use gaussian irf
    function

  * -t0 TIME_ZEROS [TIME_ZEROS ...], --time_zeros TIME_ZEROS
    [TIME_ZEROS ...] time zeros for each tscan

  * -t0f TIME_ZEROS_FILE, --time_zeros_file TIME_ZEROS_FILE filename
    for time zeros of each tscan

  * --tau [TAU [TAU ...]] lifetime of each component

  * --no_base             exclude baseline for fitting

  * -o OUT, --out OUT     prefix for output files

**Parameter bound scheme**

* fwhm: temporal width of x-ray pulse

  * lower bound: 0.5*fwhm_init

  * upper bound: 2*fwhm_init

* t_0: timezero for each scan

  * lower bound: t_0 - 2*fwhm_init

  * upper bound: t_0 + 2*fwhm_init

* tau: life_time of each component

  * if tau < 0.1

    * lower bound: tau/2

    * upper bound: 1

  * if 0.1 < tau < 10

    * lower bound: 0.05

    * upper bound: 100

  * if 10 < tau < 100

    * lower bound: 5

    * upper bound: 500

  * if 100 < tau < 1000

    * lower bound: 50

    * upper bound: 2000

  * if 1000 < tau then

    * lower bound: tau/2

    * upper bound: np.inf

**Mixing parameter eta**

For pseudo voigt IRF function, mixing parameter eta is guessed to

   \eta =
   1.36603({fwhm}_L/f)-0.47719({fwhm}_L/f)^2+0.11116({fwhm}_L/f)^3

where

   \begin{align*} f &=
   ({fwhm}_G^5+2.69269{fwhm}_G^4{fwhm}_L+2.42843{fwhm}_G^3{fwhm}_L^2
   \\   &+ 4.47163{fwhm}_G^2{fwhm}_L^3+0.07842{fwhm}_G{fwhm}_L^4 \\
   &+ {fwhm}_L^5)^{1/5} \end{align*}

This guess is according to J. Appl. Cryst. (2000). **33**, 1311-1316
'''
 
doc['TRXASprefitpack_info'] = '''
TRXASprefitpack_info
********************

Access the document of TRXASprefitpack

Type "TRXASprefitpack_info func_name" to get a description of
func_name
'''
 
doc['auto_scale'] = '''
auto_scale
**********

Automatic scaling escan and tscan data using "A-method"

Note:

  auto_scale assume each escan have same energy range and and each
  tscan have same time range. Also it assumes energy unit of escan
  data is KeV but assumes energy unit of tscan_energy_file is eV.
  Moreover energy unit of scaled_escan is eV. However time unit for
  tscan data and escan time must be same.

* usage: auto_scale    [-h] [-p PREFIX] [-ne NUM_OF_ESCAN] [-et
  ESCAN_TIME] [-re REF_ESCAN_INDEX] [-nt NUM_OF_TSCAN] [-te
  TSCAN_ENERGY] [-t0 TIME_ZEROS] [-ti TSCAN_INDEX_TO_SCALE
  [TSCAN_INDEX_TO_SCALE ...]] [-a PARM_A] {-1,0,1,2,3,4}

* positional arguments:

  * {-1,0,1,2,3,4}        current stage, set stage to -1 get detailed
    description

* optional arguments:

  * -h, --help            show this help message and exit

  * -p PREFIX, --prefix PREFIX prefix for both escan and tscan file,
    it will read prefix_escan_i.txt and prefix_tscan_j.txt

  * -ne NUM_OF_ESCAN, --num_of_escan NUM_OF_ESCAN the number of escan
    files

  * -et ESCAN_TIME, --escan_time ESCAN_TIME filename for escan delay
    times (unit: ps)

  * -re REF_ESCAN_INDEX, --ref_escan_index REF_ESCAN_INDEX index of
    escan used to the reference for scaling

  * -nt NUM_OF_TSCAN, --num_of_tscan NUM_OF_TSCAN the number of tscan
    files

  * -te TSCAN_ENERGY, --tscan_energy TSCAN_ENERGY filename for tscan
    energy (unit: eV)

  * -t0 TIME_ZEROS, --time_zeros TIME_ZEROS filename for time zero of
    each tscan (unit: ps)

  * -ti TSCAN_INDEX_TO_SCALE [TSCAN_INDEX_TO_SCALE ...],
    --tscan_index_to_scale TSCAN_INDEX_TO_SCALE [TSCAN_INDEX_TO_SCALE
    ...] tscan index to scale, use blank separation for multiple
    arguments

  * -a PARM_A, --parm_A PARM_A filename for the parameter A obtained
    from fit_static

* Stage discription

* stage -1: description It prints the description about each stages
  and aborts.

  Note:

    For every stage except -1, requires prefix, num_of_escan,
    escan_time, num_of_tscan, tscan_energy, time_zeros

* stage 0: init scaling

  Additionally requires: ref_escan_index, tscan_index_to_scale, parm_A

  the program read escan_data from
  prefix_escan_1.txt,...,prefix_num_escan.txt and tscan_data from
  prefix_tscan_1.txt,...,prefix_tscan_num_tscan.txt Also, it read file
  for parameter A generated by fit static. Then it fits scaling of
  escan_data and tscan_i_1,...,tscan_i_N to escan_e_ref. Now it
  generates prefix_escan_scaled.txt (energy unit: eV)
  prefix_escan_eps_scaled.txt prefix_tscan_scaled.txt
  prefix_tscan_eps_scaled.txt prefix_A_ref.txt If you do not include
  tscan j for scaling. You can see (j+1) th column of
  prefix_tscan_scaled.txt and j th column of tscan_scaled_eps.txt are
  filled with zeros.

* stage 1: Correction

  Additionally requires: ref_escan_index, tscan_index_to_scale

  Note:

    ref_escan_index and tscan_index_to_scale must be set to same as
    stage 0

  the program read scaled escan data and tscan data from
  prefix_escan_scaled.txt, prefix_escan_eps_scaled.txt,
  prefix_tscan_scaled.txt, prefix_tscan_eps_scaled.txt, and then it
  corrects scaling of escan using tscan_i_1 It regenerates
  prefix_escan_scaled.txt and prefix_escan_eps_scaled.txt

* stage 2: further scaling

  Additionally requires: tscan_index_to_scale

  Note:

    in this stage do not need to give the file name for parameter A.

  the program read scaled escan data and tscan data from
  prefix_escan_scaled.txt, prefix_escan_eps_scaled.txt,
  prefix_tscan_scaled.txt, prefix_tscan_eps_scaled.txt, and then it
  fits scaling of tscan_i'_1,...,tscan_i'*N' to escan_e_ref'. (Prime
  means i_1,...,i_N and e_ref values are different from stage 0) Then
  it regenerates all prefix**.txt except prefix_A_ref.txt

* stage 3: sanity check

  In this stage, the program assume, every tscan data are scaled to
  escan data. For sanity check, it gives a graph for you.

* stage 4: Scaling with another tscan data set Scale Another set of
  tscan data using already scaled escan data

  Additionally requires: tscan_index_to_scale

  Note:

    Before proceed stage 4, you should move prefix_tscan_scaled.txt,
    prefix_tscan_eps_scaled.txt and prefix_tscan_xxx.txt to some
    backup folder and rename your another tscan data set to
    prefix_tscan_xxx.txt Also you should give time_zero and energy for
    such tscan.

  In this stage, it reads scaled escan data and eps from
  prefix_escan_scaled.txt and prefix_escan_eps_scaled.txt Then it
  reads unscaled tscan data and eps from prefix_tscan_*.txt Next, it
  scales tscan datas just like stage 1. After stage4 finished you
  should go to stage 2 and stage 3.

* A-method

  When we measure each escan, we measure static spectrum (flu_off) to
  get difference spectrum (flu_on - flu_off) Since we have theortical
  static spectrum, we can fit theortical spectrum with measured static
  spectrum using following model.

     y = A \cdot {conv}({spec}_{thy}, {voigt}({fwhm}_G, {fwhm}_L,
     {peakshift})) + {baseline}

  Also, using "fit_static", you can simultaneous fitting each escan.
  During fitting process, it assume fwhm_G, fwhm_L, peak_shift are
  same in each escan, only scaling factor A and base_line are
  different. Parameter base_line reflects environmental effects, so
  pure static signal(w/o environmental effects) is

     A \cdot {conv}({spec}_{thy}, {voigt}({fwhm}_G, {fwhm}_L,
     {peakshift}))

  Each escan has same fwhm_G, fwhm_L and peak_shift value, so
  Parameter A (scaling parameter) could tell relative scaling of each
  escan. In other words, we can fit scaling of escan data just
  multipling A_ref/A to each escan. However, due to inconsistence of
  laser overlap between energy scans, one good time delay scan is
  needed to correct the scaling of energy scans.
'''
 
doc['broadening'] = '''
broadening
**********

voigt broaden theoritical calculated lineshape spectrum

* usage: broadening [-h] [-o OUT] peak e_min e_max A fwhm_G fwhm_L
  peak_shift

* positional arguments:

  * peak               filename for calculated line shape spectrum

  * e_min              minimum energy

  * e_max              maximum energy

  * A                  scale factor

  * fwhm_G             Full Width at Half Maximum of gaussian shape

  * fwhm_L             Full Width at Half Maximum of lorenzian shape

  * peak_shift         discrepancy of peak position between theory and
    experiment

* optional arguments:

  * -h, --help         show this help message and exit

  * -o OUT, --out OUT  prefix for output files
'''
 
__info__.appnd('version', doc['version']) 
__info__.appnd('description', doc['description']) 
__info__.appnd('licence', doc['licence']) 
__info__.appnd('lgpl-3.0', doc['lgpl-3.0']) 
__info__.appnd('gen_theory_data', doc['gen_theory_data']) 
__info__.appnd('automate_scaling', doc['automate_scaling']) 
__info__.appnd('corr_a_method', doc['corr_a_method']) 
__info__.appnd('exp_conv_gau', doc['exp_conv_gau']) 
__info__.appnd('exp_conv_cauchy', doc['exp_conv_cauchy']) 
__info__.appnd('solve_model', doc['solve_model']) 
__info__.appnd('compute_model', doc['compute_model']) 
__info__.appnd('compute_signal_gau', doc['compute_signal_gau']) 
__info__.appnd('compute_signal_cauchy', doc['compute_signal_cauchy']) 
__info__.appnd('compute_signal_pvoigt', doc['compute_signal_pvoigt']) 
__info__.appnd('model_n_comp_conv', doc['model_n_comp_conv']) 
__info__.appnd('fact_anal_exp_conv', doc['fact_anal_exp_conv']) 
__info__.appnd('fit_static', doc['fit_static']) 
__info__.appnd('fit_tscan', doc['fit_tscan']) 
__info__.appnd('TRXASprefitpack_info', doc['TRXASprefitpack_info']) 
__info__.appnd('auto_scale', doc['auto_scale']) 
__info__.appnd('broadening', doc['broadening']) 
__info__ = __info__.info
del doc


'''
TRXASprefitpack: 
pre fitting package for TRXAS which aims for the first order dynamics
TRXAS stands for Time Resolved X-ray Absorption Spectroscopy

:copyright: 2021 by pistack (Junho Lee)
:license: LGPL3.
'''

from .thy import *
from .data_process import *
from .mathfun import *
from .doc import *



# thy.py
# Author: Pistack
# Date: 2021. 08. 29.
# Email: pistatex@yonsei.ac.kr
# Part of XAS-pre-fit-pack package
# Useful function for broeadening spectrum

import numpy as np
from scipy.special import voigt_profile 


def gen_theory_data(e, peaks, A, fwhm_G, fwhm_L, peak_shift, out=None):

    '''
    voigt broadening theoretically calculated lineshape spectrum
    if out is not none:
    It will make
    out_thy.txt: txt file for rescaled and boroadend calc spectrum
    out_thy_stk.txt: txt file for rescaled and shifted calc peaks


    :param numpy_1d_array e: energy (unit: eV)
    :param float A: scaling parameter
    :param float fwhm_G: 
     full width at half maximum of gaussian shape (unit: eV)
    :param float fwhm_L: 
     full width at half maximum of lorenzian shape (unit: eV)
    :param float peak_shift: 
     discrepency of peak position between expt data and theoretically 
     broadened spectrum
    :param string out: prefix for output txt file [optional]

    
    :return: voigt broadened calc spectrum
    :rtype: numpy_1d_array
    '''

    num_e = e.shape[0]
    num_peaks = peaks.shape[0]
    V_matrix = np.zeros((num_e, num_peaks))

    for i in range(num_peaks):
        V_matrix[:, i] = voigt_profile(e-(peaks[i, 0]-peak_shift),
                                       fwhm_G/(2*np.sqrt(2*np.log(2))),
                                       fwhm_L/2)

    broadened_theory = A * V_matrix @ peaks[:, 1].reshape((num_peaks, 1))
    broadened_theory = broadened_theory.flatten()

    if out is not None:
        save_thy = np.vstack((e, broadened_theory))
        np.savetxt(out+'_thy'+'.txt', save_thy.T, fmt=['%.2f', '%.8e'],
                   header='energy \t abs_thy', newline='\n', delimiter='\t')
        np.savetxt(out+'_stk_rescaled'+'.txt',
                   np.vstack((peaks[:, 0]-peak_shift, A*peaks[:, 1])).T,
                   fmt=['%.2f', '%.8e'], header='energy \t abs_thy',
                   newline='\n', delimiter='\t')

    return broadened_theory
